import logging

import snowflake.connector

from workflow import constants
from workflow.utils import call_with_overhead
from workflow.common import SnowflakeConfig, RaiConfig


def begin_data_sync(logger: logging.Logger, snowflake_config: SnowflakeConfig, rai_config: RaiConfig, resources, src):
    conn = __get_connection(snowflake_config)
    cursor = conn.cursor()
    logger = logger.getChild("snowflake")

    destination_rel = src['source']
    source_table = resources[0]['uri']
    database = rai_config.database
    engine = rai_config.engine
    # Start data stream commands
    commands = (
        f"CALL RAI.use_rai_database('{database}');",
        f"CALL RAI.use_rai_engine('{engine}');",
        f"CALL RAI.create_data_stream('{source_table}', '{database}', 'simple_source_catalog, :{destination_rel}');"
    )
    try:
        for command in commands:
            logger.info(f"Executing command Snowflake command: `{command.strip()}`")
            cursor.execute(command)
    except Exception as e:
        cursor.close()
        conn.close()
        raise e

async def await_data_sync(logger: logging.Logger, snowflake_config: SnowflakeConfig, resources):
    source_table = resources[0]['uri']
    conn = __get_connection(snowflake_config)
    cursor = conn.cursor()
    logger = logger.getChild("snowflake")
    # Wait for data sync finish
    try:
        logger.info(f"Wait for Snowflake data sync finish for `{source_table}`...")
        await call_with_overhead(
            f=lambda: sync_finished(logger, cursor, source_table),
            logger=logger,
            overhead_rate=0.5,
            timeout=30 * 60,  # 30 min
            first_delay=10,  # 10 sec since it can take some time to start Job on Snowflake by Ingestion Service
            max_delay=55  # 55 sec since snowflake warehouse can be suspended after 60 sec of inactivity
        )
    except Exception as e:
        raise e
    finally:
        # Clean up data stream after sync
        cursor.execute(f"CALL RAI.delete_data_stream('{source_table}')")
        cursor.close()
        conn.close()


def sync_finished(logger: logging.Logger, cursor, source_table: str):
    cursor.execute(f"CALL RAI.get_data_stream_status('{source_table}');")
    properties = cursor.fetchall()
    key_value_pairs = {stream[0]: stream[1] for stream in properties}
    health_status = key_value_pairs.get(constants.SNOWFLAKE_STREAM_HEALTH_STATUS)
    if health_status != constants.SNOWFLAKE_HEALTHY_STREAM_STATUS:
        raise Exception(f"Snowflake sync for `{source_table}` has failed. Health status: {health_status}")
    if key_value_pairs.get(constants.SNOWFLAKE_SYNC_STATUS) == constants.SNOWFLAKE_FINISHED_SYNC_STATUS:
        synced_rows = key_value_pairs.get(constants.SNOWFLAKE_TOTAL_ROWS)
        logger.info(f"Snowflake sync finished for `{source_table}` has finished. Synced row: {synced_rows}")
        return True
    return False


def __get_connection(config: SnowflakeConfig):
    """
    This function will get a connection to snowflake.
    """
    # connect to snowflake
    return snowflake.connector.connect(
        user=config.user,
        password=config.password,
        account=config.account,
        role=config.role,
        warehouse=config.warehouse,
        database=config.database,
        schema=config.schema,
    )



import asyncio
import logging
import os
import time

from datetime import datetime, timedelta
from typing import List, Dict

from workflow import constants
from workflow.common import LocalConfig


def range_days(start: datetime, end: datetime) -> List[datetime]:
    delta = end - start
    return [start + timedelta(days=i) for i in range(delta.days + 1)]


def parse_date(date_str: str) -> datetime:
    return datetime.strptime(date_str, constants.DATE_FORMAT)


def get_date_path(end_date: str, date_format: str, day_offset: int):
    date = datetime.strptime(end_date, date_format) - timedelta(days=day_offset)
    return f"{constants.DATE_PREFIX}{date.strftime(date_format)}"


def read(fname: str) -> str:
    with open(fname) as fp:
        return fp.read()


def sansext(fname: str) -> str:
    return os.path.splitext(os.path.basename(fname))[0]


def save_csv_output(outputs: Dict, config: LocalConfig) -> None:
    """
    Save the content of dictionary as CSV files
    :param outputs:     dictionary with outputs
    :param config:      local config
    :return:
    """
    for output in outputs.keys():
        # for the time being, this handles the specialized relations of meta-exports
        normalized_file_name = output.replace("/:", "_")
        with open(f"{config.data_path}/{normalized_file_name}.csv", "w") as file:
            file.write(outputs[output])


def build_relation_path(relation: str, *keys: str) -> str:
    """
    Build relation from base and paths.
    Ex. base:`batch:config`, paths: ['daily', 'fake'] => `batch:config:daily:fake`
    :param relation:    base relation
    :param keys:        path parts
    :return: new relation
    """
    relation_path = ":" + ":".join(keys) if len(keys) > 0 else ""
    return f"{relation}{relation_path}"


def build_models(filenames: List[str], files_root: str) -> dict:
    """
    Build RAI models from list of files.
    :param filenames:   file names
    :param files_root:  files root path
    :return:
    """
    models = {}
    for file in filenames:
        filename = f"{files_root}/{file}"
        with open(filename) as fp:
            models[file] = fp.read()
    return models


def format_duration(seconds: float) -> str:
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    hours_int = int(hours)
    minutes_int = int(minutes)
    seconds_int = int(seconds)
    if hours_int != 0:
        return f"[{hours_int:d}h {minutes_int:d}m {seconds_int:d}s]"
    elif minutes_int != 0:
        return f"[{minutes_int:d}m {seconds_int:d}s]"
    else:
        return f"[{seconds_int:d}s]"


def get_common_model_relative_path(file) -> str:
    """
    Get relative path to common model from folder of given file.
    :param file:    file
    :return:
    """
    return os.path.dirname(os.path.realpath(file)) + constants.COMMON_MODEL_RELATIVE_PATH


def extract_date_range(logger: logging.Logger, start_date_str, end_date_str, number_of_days, offset_of_days) ->\
        List[str]:
    end_date = parse_date(end_date_str)
    offset = offset_of_days if offset_of_days is not None else 0
    end_date = end_date - timedelta(offset)
    start_date_adjusted = end_date - timedelta(number_of_days - 1) if number_of_days is not None else None
    if start_date_str:
        start_date = parse_date(start_date_str)
        start_date = start_date_adjusted if start_date_adjusted is not None and start_date < start_date_adjusted\
            else start_date
    else:
        start_date = start_date_adjusted
    start_date = start_date if start_date is not None else end_date
    logger.info(f"Building range from '{start_date}' to '{end_date}'")
    return [date.strftime(constants.DATE_FORMAT) for date in range_days(start_date, end_date)]


def to_rai_date_format(date_format: str) -> str:
    fmt_part_map = {
        "%Y": "YYYY",
        "%m": "mm",
        "%d": "dd"
    }
    rai_date_format = date_format
    for py_fmt, rai_fmt in fmt_part_map.items():
        rai_date_format = rai_date_format.replace(py_fmt, rai_fmt)
    return rai_date_format


async def call_with_overhead(
        f,
        logger: logging.Logger,
        overhead_rate: float,
        start_time: int = time.time(),
        timeout: int = None,
        max_tries: int = None,
        first_delay: float = 0.5,
        max_delay: int = 120,  # 2 minutes
):
    tries = 0
    max_time = time.time() + timeout if timeout else None

    while True:
        logger.debug(f"Calling function. The number of try: {tries + 1}")
        if f():
            break

        if max_tries is not None and tries >= max_tries:
            raise Exception(f'max tries {max_tries} exhausted')

        if max_time is not None and time.time() >= max_time:
            raise Exception(f'timed out after {timeout} seconds')

        tries += 1
        duration = min((time.time() - start_time) * overhead_rate, max_delay)
        if tries == 1:
            logger.debug(f"Sleep duration for the first try: {first_delay}s")
            await asyncio.sleep(first_delay)
        else:
            logger.debug(f"Sleep duration for a try: {duration}s")
            await asyncio.sleep(duration)



import logging
from typing import List
from azure.storage.blob import BlobServiceClient

from workflow.common import FileFormat
from workflow.common import AzureConfig
from workflow.constants import BLOB_PAGE_SIZE


def list_files_in_containers(logger: logging.Logger, config: AzureConfig, path_prefix) -> List[str]:
    blob_service_client = BlobServiceClient(account_url=f"https://{config.account}.blob.core.windows.net",
                                            credential=config.sas)
    container_client = blob_service_client.get_container_client(config.container)
    logger = logger.getChild("blob")

    # Get a list of blobs in the folder
    logger.debug(f"Path prefix to list blob files: {path_prefix}")
    paths = []
    for page in container_client.list_blobs(name_starts_with=path_prefix, results_per_page=BLOB_PAGE_SIZE).by_page():
        for blob in page:
            blob_name = blob.name
            if FileFormat.is_supported(blob_name):
                paths.append(f"azure://{config.account}.blob.core.windows.net/{config.container}/{blob_name}")
            else:
                logger.debug(f"Skip unsupported file from blob: {blob_name}")
    return paths



import logging


def configure(level=logging.INFO) -> logging.Logger:
    # override default logging level for azure
    logger = logging.getLogger('azure.core')
    logger.setLevel(logging.ERROR)
    logger = logging.getLogger('snowflake')
    logger.setLevel(logging.ERROR)

    logger = logging.getLogger()
    logger.setLevel(level)
    handler = logging.StreamHandler()
    formatter = logging.Formatter('[%(asctime)s][%(levelname)s][%(name)s] %(message)s', "%H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger.getChild("cli")



import logging
import re
import json
import time
from typing import Dict, Any
from urllib.error import HTTPError
from railib import api, config

from workflow import query as q
from workflow.constants import RAI_PROFILE, RAI_PROFILE_PATH, RAI_SDK_HTTP_RETRIES
from workflow.common import RaiConfig


def get_config(engine: str, database: str, env_vars: dict[str, Any]) -> RaiConfig:
    """
    Create RAI config for given parameters
    :param engine:          RAI engine
    :param database:        RAI database
    :param env_vars:        Env vars dictionary
    :return: Env config
    """
    rai_profile = env_vars.get(RAI_PROFILE, "default")
    rai_profile_path = env_vars.get(RAI_PROFILE_PATH, "~/.rai/config")
    retries = env_vars.get(RAI_SDK_HTTP_RETRIES, 3)
    ctx = api.Context(**config.read(fname=rai_profile_path, profile=rai_profile), retries=retries)
    return RaiConfig(ctx=ctx, engine=engine, database=database)


def load_json(logger: logging.Logger, rai_config: RaiConfig, relation: str, json_data: str) -> None:
    """
    Load json into RAI DB
    :param logger:      logger
    :param rai_config:  RAI config
    :param relation:    relation for insert
    :param json_data:   json data string
    :return:
    """
    logger.info(f"Loading json as '{relation}'")
    logger.debug(f"Json content: '{json_data}'")
    query_model = q.load_json(relation, json_data)
    execute_query(logger, rai_config, query_model.query, query_model.inputs, readonly=False)


def create_engine(logger: logging.Logger, rai_config: RaiConfig, size: str = "XS") -> None:
    """
    Create RAI engine specified in RAI config.
    Waits until provision has finished.
    :param logger:      logger
    :param rai_config:  RAI config
    :param size:        RAI engine size
    :return:
    """
    logger.info(f"Creating engine `{rai_config.engine}`")
    api.create_engine_wait(rai_config.ctx, rai_config.engine, size)


def delete_engine(logger: logging.Logger, rai_config: RaiConfig) -> None:
    """
    Delete RAI engine specified in RAI config.
    :param logger:      logger
    :param rai_config:  RAI config
    :return:
    """
    logger.info(f"Deleting engine `{rai_config.engine}`")
    api.delete_engine(rai_config.ctx, rai_config.engine)
    # make sure that engine was deleted
    api.poll_with_specified_overhead(
        lambda: not engine_exist(logger, rai_config),
        overhead_rate=0.2,
        timeout=10 * 60,
    )


def engine_exist(logger: logging.Logger, rai_config: RaiConfig) -> bool:
    """
    Check if RAI engine specified in RAI config does exist on RAI Cloud.
    :param logger:      logger
    :param rai_config:  RAI config
    :return: `True` if the engine does exist otherwise `False`
    """
    logger.info(f"Check if engine `{rai_config.engine}` exist")
    return bool(api.get_engine(rai_config.ctx, rai_config.engine))


def create_database(logger: logging.Logger, rai_config: RaiConfig, source_db=None) -> None:
    """
    Create RAI DB specified in RAI config. The HTTP error 409(Conflict) swallowed since DB already exists.
    :param logger:      logger
    :param rai_config:  RAI config
    :param source_db:   source DB to clone from
    :return:
    """
    logger.info(f"Creating database `{rai_config.database}`")
    if source_db:
        logger.info(f"Use `{source_db}` database for clone")
    try:
        api.create_database(rai_config.ctx, rai_config.database, source_db)
    except HTTPError as e:
        if e.status == 409:
            logger.info(f"Database '{rai_config.database}' already exists")
        else:
            raise e


def delete_database(logger: logging.Logger, rai_config: RaiConfig) -> None:
    """
    Delete RAI DB specified in RAI config.
    :param logger:      logger
    :param rai_config:  RAI config
    :return:
    """
    logger.info(f"Deleting database `{rai_config.database}`")
    try:
        api.delete_database(rai_config.ctx, rai_config.database)
    except HTTPError as e:
        raise e


def database_exist(logger: logging.Logger, rai_config: RaiConfig) -> bool:
    """
    Check if RAI DB specified in RAI config does exist
    :param logger:      logger
    :param rai_config:  RAI config
    :return: `True` if DB exists otherwise `False`
    """
    logger.info(f"Check if db `{rai_config.database}` exist")
    return bool(api.get_database(rai_config.ctx, rai_config.database))


def install_models(logger: logging.Logger, rai_config: RaiConfig, models: dict) -> None:
    """
    Install Rel model into RAI DB specified in RAI config.
    :param logger:      logger
    :param rai_config:  RAI config
    :param models:      RAI models to install
    :return:
    """
    logger.info("Installing models")

    query_model = q.install_model(models)
    execute_query(logger, rai_config, query_model.query, query_model.inputs, False, False)


def execute_query(logger: logging.Logger, rai_config: RaiConfig, query: str, inputs: dict = None, readonly: bool = True,
                  ignore_problems: bool = False) -> api.TransactionAsyncResponse:
    """
    Execute Rel query using DB and engine from RAI config.
    :param logger:          logger
    :param rai_config:      RAI config
    :param query:           Rel query
    :param inputs:          Rel query inputs
    :param readonly:        Parameter to specify transaction type: Read/Write.
    :param ignore_problems: Ignore SDK problems if any
    :return: SDK response
    """
    try:
        logger.info(f"Execute query: database={rai_config.database} engine={rai_config.engine} readonly={readonly}")
        start_time = int(time.time())
        txn = api.exec_async(rai_config.ctx, rai_config.database, rai_config.engine, query, readonly, inputs)
        logger.info(f"Execute query: transaction id - {txn.transaction['id']}")

        # in case of if short-path, return results directly, no need to poll for state
        if not (txn.results is None):
            return txn

        logger.info(f"Execute query: polling for transaction with id - {txn.transaction['id']}")
        rsp = api.TransactionAsyncResponse()
        txn = api.get_transaction(rai_config.ctx, txn.transaction["id"])

        api.poll_with_specified_overhead(
            lambda: api.is_txn_term_state(api.get_transaction(rai_config.ctx, txn["id"])["state"]),
            overhead_rate=0.2,
            start_time=start_time
        )

        rsp.transaction = api.get_transaction(rai_config.ctx, txn["id"])
        rsp.metadata = api.get_transaction_metadata(rai_config.ctx, txn["id"])
        rsp.problems = api.get_transaction_problems(rai_config.ctx, txn["id"])
        rsp.results = api.get_transaction_results(rai_config.ctx, txn["id"])

        _assert_problems(logger, rsp, ignore_problems)
        return rsp
    except HTTPError as e:
        raise e


def execute_relation_json(logger: logging.Logger, rai_config: RaiConfig, relation: str,
                          ignore_problems: bool = False) -> Dict:
    """
    Execute Rel query with output relation as a json string and parse the output.
    :param logger:          logger
    :param rai_config:      RAI config
    :param relation:        Rel relations
    :param ignore_problems: Ignore SDK problems if any
    :return: parsed json string
    """
    rsp = execute_query(logger, rai_config, q.output_json(relation), ignore_problems=ignore_problems)
    return _parse_json_string(rsp)


def execute_query_csv(logger: logging.Logger, rai_config: RaiConfig, query: str, ignore_problems: bool = False) -> Dict:
    """
    Execute query and parse the output as CSV.
    :param logger:          logger
    :param rai_config:      RAI config
    :param query:           Rel query
    :param ignore_problems: Ignore SDK problems if any
    :return: parsed CSV output
    """
    rsp = execute_query(logger, rai_config, query, ignore_problems=ignore_problems)
    return _parse_csv_string(rsp)


def execute_query_take_single(logger: logging.Logger, rai_config: RaiConfig, query: str, readonly: bool = True,
                              ignore_problems: bool = False) -> any:
    """
    Execute query and take the first result.
    :param logger:          logger
    :param rai_config:      RAI config
    :param query:           Rel query
    :param readonly:        Parameter to specify transaction type: Read/Write
    :param ignore_problems: Ignore SDK problems if any
    """
    rsp = execute_query(logger, rai_config, query, readonly=readonly, ignore_problems=ignore_problems)
    if not rsp.results:
        logger.info(f"Query returned no results: {query}")
        return None
    return rsp.results[0]['table'].to_pydict()["v1"][0]


def _assert_problems(logger: logging.Logger, rsp: api.TransactionAsyncResponse, ignore_problems: bool):
    problems_has_error = _handle_problems(logger, rsp.problems)
    txn_state = rsp.transaction.get('state')
    if txn_state != 'COMPLETED' or (problems_has_error and not ignore_problems):
        # todo: create a hierarchy of exceptions
        raise AssertionError(
            f"Transaction did not complete successfully or has problems: {txn_state}, id: {rsp.transaction['id']}")


def _handle_problems(logger: logging.Logger, problems: list) -> bool:
    problems_has_error = False
    for problem in problems:
        if 'is_error' in problem and problem['is_error'] is True or \
                'is_exception' in problem and problem['is_exception'] is True:
            problems_has_error = True
            logger.error(problem)
        else:
            logger.warning(problem)
    return problems_has_error


def _parse_json_string(rsp: api.TransactionAsyncResponse) -> Dict:
    """
    Parse the output for a json_string query ie
        def output = json_string[...]
    """
    if not rsp.results:
        return {}
    pa_table = rsp.results[0]['table']
    data = pa_table.to_pydict()
    return json.loads(data["v1"][0])


def _parse_csv_string(rsp: api.TransactionAsyncResponse) -> Dict:
    """
    Parse the output for a csv_string query ie
        def output:{rel_name} = csv_string[...]
    """
    resp = {}
    rel_pattern = r'^/:output/:(.*)/String$'
    for result in rsp.results:
        match = re.search(rel_pattern, result['relationId'])
        if match:
            relation_id = match.group(1)
            data = result['table'].to_pydict()["v1"][0]
            resp[relation_id] = data
    return resp



import logging
import unittest
from unittest.mock import Mock, MagicMock

from workflow import paths
from workflow import constants
from workflow.common import Source, BatchConfig, EnvConfig, Container, ContainerType
from workflow.executor import ConfigureSourcesWorkflowStepFactory, WorkflowConfig, WorkflowStepState


class TestConfigureSourcesWorkflowStepFactory(unittest.TestCase):
    logger: logging.Logger = Mock()

    def test_get_step(self):
        # Setup factory spy
        factory = ConfigureSourcesWorkflowStepFactory()
        config = _create_wf_cfg(EnvConfig({"azure": _create_container("default", ContainerType.AZURE),
                                           "local": _create_container("local", ContainerType.LOCAL)}), Mock())
        spy = MagicMock(wraps=factory._parse_sources)
        sources = [_create_test_source(config.env.get_container("azure"), "src1"),
                   _create_test_source(config.env.get_container("local"), "src2"),
                   _create_test_source(config.env.get_container("local"), "src3"),
                   _create_test_source(config.env.get_container("local"), "src4")]
        spy.return_value = sources
        factory._parse_sources = spy

        # When
        step = factory._get_step(self.logger, config, "1", "name", WorkflowStepState.INIT, 0, None, {"configFiles": []})
        # Then
        self.assertEqual("1", step.idt)
        self.assertEqual("name", step.name)
        self.assertEqual(WorkflowStepState.INIT, step.state)
        self.assertEqual(0, step.timing)
        self.assertEqual(None, step.engine_size)
        self.assertEqual([], step.config_files)
        self.assertEqual("./rel", step.rel_config_dir)
        self.assertEqual(sources, step.sources)
        self.assertEqual(2, len(step.paths_builders.keys()))
        self.assertIsInstance(step.paths_builders.get("local"), paths.LocalPathsBuilder)
        self.assertIsInstance(step.paths_builders.get("default"), paths.AzurePathsBuilder)
        self.assertEqual("2021-01-01", step.start_date)
        self.assertEqual("2021-01-01", step.end_date)
        self.assertFalse(step.force_reimport)
        self.assertFalse(step.force_reimport_not_chunk_partitioned)


def _create_test_source(container: Container, relation: str) -> Source:
    return Source(
        container=container,
        relation=relation,
        relative_path="test",
        input_format="test",
        extensions=["test"],
        is_chunk_partitioned=True,
        is_date_partitioned=True,
        loads_number_of_days=0,
        offset_by_number_of_days=0,
        snapshot_validity_days=0,
        paths=[]
    )


def _create_container(name: str, c_type: ContainerType) -> Container:
    return Container(
        name=name,
        type=c_type,
        params={}
    )


def _create_wf_cfg(env_config: EnvConfig, batch_config: BatchConfig) -> WorkflowConfig:
    parameters = {
        constants.REL_CONFIG_DIR: "./rel",
        constants.START_DATE: "2021-01-01",
        constants.END_DATE: "2021-01-01",
        constants.FORCE_REIMPORT: False,
        constants.FORCE_REIMPORT_NOT_CHUNK_PARTITIONED: False,
        constants.COLLAPSE_PARTITIONS_ON_LOAD: False
    }
    return WorkflowConfig(
        env=env_config,
        batch_config=batch_config,
        recover=False,
        recover_step="",
        selected_steps=[],
        step_params=parameters
    )
